/*----------------------------------------------------------------------------*\
     mtxe - matrix engine
     version 1.0.2

     Copyright (C) 2004, 2005, 2006 Ville Jokela
     This program is free software; you can redistribute it and/or modify
     it under the terms of the GNU General Public License as published by
     the Free Software Foundation; either version 2 of the License, or
     (at your option) any later version.

     This program is distributed in the hope that it will be useful,
     but WITHOUT ANY WARRANTY; without even the implied warranty of
     MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
     GNU General Public License for more details.

     You should have received a copy of the GNU General Public License
     along with this program; if not, write to the Free Software
     Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
\*----------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
# include <config.h>
#endif

#ifdef HAVE_STDIO_H
# include <stdio.h>
#endif /* HAVE_STDIO_H */
#ifdef HAVE_STDLIB_H
# include <stdlib.h>
#endif /* HAVE_STDLIB_H */

#if HAVE_GETTIMEOFDAY
# include <sys/time.h>
#elif HAVE_FTIME
# include <sys/timeb.h>
#endif /* time */

#include "mtxe.h"

static int size_check(mtxe_t *mtxe);

mtxe_t* mtxe_init(unsigned char flags)
{
#	if HAVE_GETTIMEOFDAY
	struct timeval tv;
#	elif HAVE_FTIME
	struct timeb tb;
#	endif  /* time */
	mtxe_t* mtxe;

	mtxe = malloc(sizeof(mtxe_t));
	if (mtxe == NULL)
		return NULL;
	mtxe->col = NULL;
	mtxe->cols = 0;
	mtxe->vis.x = 0;
	mtxe->vis.y = 0;
	mtxe->units = 0;
	mtxe->flags = flags;
#	if HAVE_GETTIMEOFDAY
	if (!gettimeofday(&tv, NULL))
		mtxe_err_set(mtxe, MTXE_ERR_SYSTEM_CALL);
	mtxe->r = tv.tv_usec;
#	elif HAVE_FTIME
	ftime(&tb);
	mtxe->r = tb.millitm;
#	endif /* time */
	mrnd(&mtxe->r);
	
	mtxe->err = 0;

	return mtxe;
}

int mtxe_update(mtxe_t* mtxe)
{
	int i, r = 0;

	if (mtxe->err & MTXE_ERR_FATAL)
		return MTXE_ERR;
	r |= size_check(mtxe);
	for (i = 0; i < mtxe->cols; i++)
		if (mtxe->col[i] != NULL) {
			r |= col_update(mtxe, mtxe->col[i]);
		} else if (!(mtxe->flags & MTXE_QUIT))
			mtxe->col[i] = col_init(mtxe, i);
	if (!(mtxe->flags & MTXE_QUIT)) {
		/* the next might seem a bit weird, but it works pretty well.
		 * on a 80 chars wide screen we have 40 cols, so 40 / 5 == 8.
		 * every time the loop runs, we check if i is smaller than the
		 * random number from 0 to 7. this reduces the probability of
		 * large numbers of units being spawned at once, but doesn't
		 * render it impossible. (yes, this was invented by accident)
		 */
		for (i = 0; i < RAND(MAX(mtxe->vis.x / 5, 2)); i++)
			r |= unit_spawn(mtxe);
		if (mtxe->flags & MTXE_CHAOS)
			mtxe_chaos(mtxe);
	}
	else
		if (mtxe->units == 0)
			return MTXE_QUIT;

	return r;
}

void mtxe_free(mtxe_t* mtxe)
{
	int i;

	for (i = 0; i < mtxe->cols; i++)
		col_free(mtxe, mtxe->col[i]);
	free(mtxe->col);
	free(mtxe);
}

char* mtxe_err_get(mtxe_t *mtxe)
{
	if (mtxe->err & MTXE_ERR_UNIT_SLOT_RATIO) {
		mtxe->err ^= MTXE_ERR_UNIT_SLOT_RATIO;
		return "invalid unit/slot-ratio error";
	} else if (mtxe->err & MTXE_ERR_MEMORY_ALLOCATION) {
		mtxe->err ^= MTXE_ERR_MEMORY_ALLOCATION;
		return "memory allocation error";
	} else if (mtxe->err & MTXE_ERR_SLOT_MODIFICATION) {
		mtxe->err ^= MTXE_ERR_SLOT_MODIFICATION;
		return "slot modification error";
	} else if (mtxe->err & MTXE_ERR_INVALID_LOCATION) {
		mtxe->err ^= MTXE_ERR_INVALID_LOCATION;
		return "invalid location error";
	} else if (mtxe->err & MTXE_ERR_SYSTEM_CALL) {
		mtxe->err ^= MTXE_ERR_SYSTEM_CALL;
		return "system call failed";
	} else if (mtxe->err & MTXE_ERR_SCREEN_SIZE) {
		mtxe->err ^= MTXE_ERR_SCREEN_SIZE;
		return "invalid screen size";
	} else if (mtxe->err & MTXE_ERR_FATAL) {
		mtxe->err ^= MTXE_ERR_FATAL;
		return "fatal error";
	} else
		return NULL;
}

void mtxe_chaos(mtxe_t *mtxe)
{
	int col;
	int unit;
	int glyph;
	int i;

	if (mtxe->units == 0)
		return;
	
	mtxe->chaos = !((1.0 / (double) RAND(mtxe->units) + 1) * (double) RAND(mtxe->units));
	if (mtxe->chaos) {
		if (mtxe->cols == 0)
			return;
		col = RAND(mtxe->cols);
		if (mtxe->col[col]->units == 0)
			return;
		unit = RAND(mtxe->col[col]->units);
		if (mtxe->col[col]->unit[unit]->glyphs == 0)
			return;
		glyph = RAND(mtxe->col[col]->unit[unit]->glyphs);
		mtxe->chaos = RAND(16);
		if (mtxe->chaos < 6) {		/* 0 - 5 */
			mtxe->col[col]->unit[unit]->glyph[glyph] = -RAND(64);
			mtxe->col[col]->unit[unit]->flags |= UNIT_RAND;
		} else if (mtxe->chaos < 10) {	/* 6 - 9 */
			/* first we fill in the "undecided" glyphs */
			for (i = 0; i < mtxe->col[col]->unit[unit]->glyphs; i++)
				if (mtxe->col[col]->unit[unit]->glyph[i] == -1)
					mtxe->col[col]->unit[unit]->glyph[i] = glyph_rand(mtxe);
			mtxe->col[col]->unit[unit]->flags |= UNIT_SSTOP;
		} else if (mtxe->chaos < 14) {	/* 10 - 13 */
			if (mtxe->col[col]->units == 1)
				mtxe->col[col]->unit[unit]->flags |= UNIT_USTOP;
		} else {			/* 14 - 15 */
			for (i = 0; i < mtxe->col[col]->unit[unit]->glyphs; i++)
				if (mtxe->col[col]->unit[unit]->glyph[i] == -1)
					mtxe->col[col]->unit[unit]->glyph[i] = glyph_rand(mtxe);
			mtxe->col[col]->unit[unit]->flags |= UNIT_SSTOP;
			mtxe->col[col]->unit[unit]->flags |= UNIT_USTOP;
		}
	}
}

col_t* col_init(mtxe_t *mtxe, int loc)
{
	col_t* col;

	col = (col_t *) malloc(sizeof(col_t));
	if (col == NULL) {
		mtxe_err_set(mtxe, MTXE_ERR_MEMORY_ALLOCATION);
		return NULL;
	}
	col->unit = NULL;
	col->units = 0;
	col->slots = 0;
	col->loc = loc;

	return col;
}

int col_update(mtxe_t *mtxe, col_t *col)
{
	int i, r;

	for (i = r = 0; i < col->units; i++)
		r |= unit_update(mtxe, col->unit[i]);

	for (i = 0; i < col->units; i++)
		if (col->unit[i]->flags & UNIT_DEAD)
			r |= unit_free(mtxe, col->unit[i]);

	return r;
}

int col_slot_add(mtxe_t *mtxe, col_t* col)
{
	void *tmp;

	if (col->slots == col->units) {
		if (col->slots >= 0) {
			tmp = realloc(col->unit, sizeof(unit_t*) * (col->slots + 1));
			if (!tmp) {	/* col->slots + 1 shouldn't be 0 so we deserve the */
					/* memory error here */
				mtxe_err_set(mtxe, MTXE_ERR_MEMORY_ALLOCATION);
				return MTXE_ERR;
			}
			col->unit = tmp;
		} else { /* col->slots < 0, which is an error */
			mtxe_err_set(mtxe, MTXE_ERR_UNIT_SLOT_RATIO);
			return MTXE_ERR;
		}
		col->unit[col->slots] = NULL;
		col->slots++;
	} else if (col->slots > col->units) {
		mtxe_err_set(mtxe, MTXE_ERR_SLOT_MODIFICATION);
		return MTXE_ERR;
	} else { /* col->slots < col->units */
		mtxe_err_set(mtxe, MTXE_ERR_UNIT_SLOT_RATIO);
		return MTXE_ERR;
	}

	return 0;
}

int col_slot_rmv(mtxe_t* mtxe, col_t* col)
{
	void *tmp;

	if (col->slots > col->units) {
		if (col_clean(mtxe, col) == MTXE_ERR)
			return MTXE_ERR;
		if (col->slots >= 1) {
			tmp = realloc(col->unit, sizeof(unit_t*) * (col->slots - 1));
			if (tmp == NULL && sizeof(unit_t*) * (col->slots - 1) != 0) {
				mtxe_err_set(mtxe, MTXE_ERR_MEMORY_ALLOCATION);
				return MTXE_ERR;
			}
			col->unit = tmp;
		} else { /* col->slots < 1, which is an error.
			  * either in col or in calling this function */
			mtxe_err_set(mtxe, MTXE_ERR_UNIT_SLOT_RATIO);
			return MTXE_ERR;
		}
		col->slots--;
	} else if (col->slots == col->units) {
		mtxe_err_set(mtxe, MTXE_ERR_SLOT_MODIFICATION);
		return MTXE_ERR;
	} else { /* col->slots < col->units */
		mtxe_err_set(mtxe, MTXE_ERR_UNIT_SLOT_RATIO);
		return MTXE_ERR;
	}

	return 0;
}

int col_clean(mtxe_t* mtxe, col_t* col)
{
	int i, j;
	unit_t **tmp;

	if (col->slots > col->units) {
		tmp = malloc(sizeof(unit_t*) * col->slots);
		if (tmp == NULL) {
			mtxe_err_set(mtxe, MTXE_ERR_MEMORY_ALLOCATION);
			return MTXE_ERR;
		}
		for (i = 0; i < col->slots; i++)
			tmp[i] = NULL;
		for (i = 0, j = 0; i < col->slots; i++)
			if (col->unit[i] != NULL) {
				tmp[j] = col->unit[i];
				tmp[j]->loc.y = j;
				j++;
			}
		free(col->unit);
		col->unit = tmp;
	} else if (col->slots < col->units) {
		mtxe_err_set(mtxe, MTXE_ERR_UNIT_SLOT_RATIO);
		return MTXE_ERR;
	}
	/* else (col->slots == col->units) so there is no possibility of
	 * empty spaces, so we do nothing */
	return 0;
}

int col_free(mtxe_t* mtxe, col_t* col)
{
	int i, r = 0;

	/* if there are any units on this column, we free them */
	for (i = 0; i < col->units; i++)
		r |= unit_free(mtxe, col->unit[i]);
	mtxe->col[col->loc] = NULL;
	col->loc = 0;
	col->units = 0;
	col->slots = 0;
	if (col->unit != NULL)
		free(col->unit);
	free(col);

	return r;
}

unit_t* unit_init(mtxe_t* mtxe, col_t* col)
{
	unit_t* unit;
	int i;

	unit = (unit_t*) malloc(sizeof(unit_t));
	if (unit == NULL) {
		mtxe_err_set(mtxe, MTXE_ERR_MEMORY_ALLOCATION);
		return NULL;
	}
	unit->glyphs = (RAND(MTXE_GLYPHS_MAX - MTXE_GLYPHS_MIN) + MTXE_GLYPHS_MIN);
	unit->glyph = (int*) malloc(sizeof(int) * unit->glyphs);
	if (unit->glyph == NULL) {
		mtxe_err_set(mtxe, MTXE_ERR_MEMORY_ALLOCATION);
		free(unit);
		return NULL;
	}
	for (i = 0; i < unit->glyphs; i++)
		unit->glyph[i] = -1;
	unit->loc.y = col->units;
	unit->loc.x = unit->pos.x = col->loc;
	unit->pos.y = 0;
	unit->index = 0;
	unit->flags = 0;
	if (mtxe->flags & MTXE_UGLY) {
		for (i = 0; i < unit->glyphs; i++)
			unit->glyph[i] = glyph_rand(mtxe);
		unit->flags |= UNIT_OLD;
	} else
		unit->glyph[0] = glyph_rand(mtxe);

	return unit;
}

int unit_spawn(mtxe_t* mtxe)
{
	int i;
	col_t* col;

	/* choose the target column for spawning */
	do	col = mtxe->col[RAND(mtxe->vis.x)];
	while (col == NULL);

	/* check if there's space in the top of the column and that none of the units are stopped */
	for (i = 0; i < col->slots; i++) {
		if (col->unit[i] != NULL) {
			if (!((col->unit[i]->pos.y - col->unit[i]->glyphs) > 0))
				return 0;
			if (col->unit[i]->flags & UNIT_USTOP)
				return 0;
		}
	}
	if (col->units == col->slots)
		if (col_slot_add(mtxe, col) == MTXE_ERR)
			return MTXE_ERR;
	col->unit[col->slots - 1] = unit_init(mtxe, col);
	if (col->unit[col->slots - 1] == NULL) {
		col_slot_rmv(mtxe, col);
		return MTXE_ERR;
	}
	col->units++;
	mtxe->units++;

	return 0;
}

int unit_update(mtxe_t* mtxe, unit_t* unit)
{
	int i;

	if (unit->pos.y < 0 || unit->pos.x < 0) {
		mtxe_err_set(mtxe, MTXE_ERR_INVALID_LOCATION);
		return MTXE_ERR;
	}
	if (unit->pos.y >= (mtxe->vis.y + unit->glyphs))	/* it's dead */
		unit->flags |= UNIT_DEAD;		/* mark it as such */
	else { /* it's still alive */
		if (!(unit->flags & UNIT_SSTOP || unit->flags & UNIT_OLD)) {
			if (unit->index > 0)
				unit->index--;
			else
				unit->index = unit->glyphs - 1;
			if (unit->pos.y < mtxe->vis.y)
				unit->glyph[unit->index] = glyph_rand(mtxe);
			else
				unit->glyph[unit->index] = -1;
		}
		if (!(unit->flags & UNIT_USTOP))
			unit->pos.y++;
		if (unit->flags & UNIT_CHAOS) {
			if (unit->flags & UNIT_RAND) {
				for (i = 0; i < unit->glyphs; i++)
					if (unit->glyph[i] < 0) {
						if (RAND(-unit->glyph[i]))
							unit->glyph[i]++;
						else {
							unit->glyph[i] = glyph_rand(mtxe);
							unit->flags ^= UNIT_RAND;
						}
					}
			} else if (unit->flags & UNIT_SSTOP) {
				if (!RAND(8)) {
					unit->flags ^= UNIT_SSTOP;
				}
			} else if (unit->flags & UNIT_USTOP) {
				if (!RAND(8))
					unit->flags ^= UNIT_USTOP;
			}
		}
	}

	return 0;
}

int unit_free(mtxe_t* mtxe, unit_t* unit)
{
	int r = 0;
	int x;

	/* first, free the array of glyphs */
	free(unit->glyph);
	unit->glyph = NULL;
	unit->glyphs = 0;
	/* then remove the unit from the column */
	mtxe->col[unit->loc.x]->unit[unit->loc.y] = NULL;
	mtxe->col[unit->loc.x]->units--;
	/* then reset the unit */
	x = unit->loc.x;
	unit->loc.x = unit->loc.y = 0;
	unit->index = 0;
	unit->flags = 0;
	/* then free the unit */
	free(unit);
	/* then remove the slot */
	if (col_slot_rmv(mtxe, mtxe->col[x]))
		r = MTXE_ERR;
	/* then update the variables */
	mtxe->units--;

	return r;
}

int glyph_rand(mtxe_t* mtxe)
{
        if (mtxe->flags & MTXE_GLYPH)
		return RAND(56);
	else
		return RAND(71);
}

int mtxe_get_glyph(unit_t* unit, int n)
{
	if ((unit->index + n) < unit->glyphs)
		return unit->glyph[unit->index + n];
	else
		return unit->glyph[(unit->index + n) - unit->glyphs];
}

static int size_check(mtxe_t* mtxe)
{
	int i;
	col_t** tmp;

	if (mtxe->vis.x == mtxe->cols)
		return 0;
	else if (mtxe->vis.x > mtxe->cols) {
		if (mtxe->cols > 0) {
			tmp = realloc(mtxe->col, sizeof(col_t*) * mtxe->vis.x);
			if (tmp == NULL) {
				mtxe_err_set(mtxe, MTXE_ERR_MEMORY_ALLOCATION);
				return MTXE_ERR;
			} else {
				mtxe->col = tmp;
				tmp = NULL;
			}
		} else if (mtxe->cols == 0) {
			mtxe->col = malloc(sizeof(col_t*) * mtxe->vis.x);
			if (mtxe->col == NULL) {
				mtxe_err_set(mtxe, MTXE_ERR_MEMORY_ALLOCATION);
				return MTXE_ERR;
			}
		} else { /* mtxe->cols < 0 */
			mtxe_err_set(mtxe, MTXE_ERR_SCREEN_SIZE);
			return MTXE_ERR;
		}
		for (i = mtxe->cols; i < mtxe->vis.x; i++)
			mtxe->col[i] = col_init(mtxe, i);
		mtxe->cols = mtxe->vis.x;
	} else if (mtxe->vis.x < mtxe->cols) {
		/* check that none of the col's have units */
		for (i = mtxe->vis.x; i < mtxe->cols; i++)
			if (mtxe->col[i]->units != 0)	/* if they do */
				return 0;		/* then return */
		for (i = mtxe->vis.x; i < mtxe->cols; i++) {
			col_free(mtxe, mtxe->col[i]);
			mtxe->col[i] = NULL;
		}
		tmp = realloc(mtxe->col, sizeof(col_t*) * mtxe->vis.x);
		if (tmp == NULL) {
			mtxe_err_set(mtxe, MTXE_ERR_MEMORY_ALLOCATION);
			return MTXE_ERR;
		} else {
			mtxe->col = tmp;
			tmp = NULL;
		}
		mtxe->cols = mtxe->vis.x;
	}

	return 0;
}

int mrnd(unsigned long *seed)
{
	*seed = *seed * 1103515245 + 12345;

	return ((int)(*seed / 65536) % 32768);
}
